/*
 *  Copyright (C) 2008 Florent Bedoiseau (electronique.fb@free.fr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ZXSpectrum.h"
#include "Z80.h"
#include "ZXMenu.h"
#include "ZXKeyboard.h"
#include "ZXSequencer.h"
#include "ZXFiles.h"
#include <iostream>
#include <string>
#include <sys/unistd.h>
#include "bip.h"
using namespace std;

//
// MEMORY
//
ZXMemory::ZXMemory (void) {
  memory = new u8 [65536];
  for (int i = ZX_VIDEO_START; i < ZX_ATTRIBUTE_END; i++) memory [i] = PA_Rand();
}

ZXMemory::~ZXMemory (void) {
  if (memory) {
    delete [] memory;
    memory = 0;
  }
}

bool ZXMemory::LoadFromFile (const char* aFile) {
  FILE *fp = fopen (aFile, "rb");
  if (!fp) {
    return false;
  }
  
  u8 data;
  int address = 0;
  while (fread (&data, sizeof (u8), 1, fp) != 0) {
	if (address>=65536) {
	  fclose (fp);
      return 1;
	}

    WriteToZXMemory ((u16) address, data);
    address++;
  }
  
  fclose (fp);
  return true;
}

bool ZXMemory::LoadFromBin (const u8* aBin) {
  u16 address;
  for (address = 0; address < 16384; address++) {
    WriteToZXMemory ((u16) address, (u8) aBin[address]);
  }
  
  return true;
}

bool ZXMemory::SaveToFile (const char* aFile) {
  FILE *fp = fopen (aFile, "wb");
  if (!fp) {
    return false;
  }
  
  fwrite (memory, 1, 65536, fp);
  
  fclose (fp);
  return true;
}

//
// ZXULA
//

ZXULA::ZXULA (ZXMemory *aZXMemory, ZXIO *aZXIO) {
  theZXMemory = aZXMemory;
  theZXIO = aZXIO;

  // Not bright colors
  ColorTable [0] = PA_RGB (0,0,0); // Black
  ColorTable [1] = PA_RGB (0,0,28); // Blue
  ColorTable [2] = PA_RGB (28,0,0); // Red
  ColorTable [3] = PA_RGB (28,0,28); // Magenta
  ColorTable [4] = PA_RGB (0,28, 0); // Green
  ColorTable [5] = PA_RGB (0,28, 28);// Cyan
  ColorTable [6] = PA_RGB (28,28, 0); // Yellow
  ColorTable [7] = PA_RGB (28,28,28); // White

  // Bright colors
  ColorTable [8] = PA_RGB (0,0,0); // Black
  ColorTable [9] = PA_RGB (0,0,31); // Blue
  ColorTable [10] = PA_RGB (31,0,0); // Red
  ColorTable [11] = PA_RGB (31,0,31); // Magenta
  ColorTable [12] = PA_RGB (0,31, 0); // Green
  ColorTable [13] = PA_RGB (0,31, 31);// Cyan
  ColorTable [14] = PA_RGB (31,31, 0); // Yellow
  ColorTable [15] = PA_RGB (31,31,31); // White

  flash = 0;
  flash_inv = false;

  _joystick = J_KEMPSTON;

  _sound = true;
  _lastSampleValue = 0; // Last sound sample value
  theSoundBuffer = new s8 [256];
}


ZXULA::~ZXULA (void) {
  if (theSoundBuffer) {
	delete theSoundBuffer;
	theSoundBuffer = 0;
  }
}

void ZXULA::changeJoystick () {
  switch (_joystick) {
  case J_KEMPSTON: 
	 _joystick = J_SINCLAIR_1;
    break;
  case J_SINCLAIR_1: 
	_joystick = J_SINCLAIR_2;
    break;
  case J_SINCLAIR_2: 
	_joystick = J_CURSOR;
    break;
  case J_CURSOR: 
	_joystick = J_KEMPSTON;
    break;
  default:
   break;
  }

  displayJoystick ();
}

void ZXULA::setJoystick (ZX_JOYSTICK joy) {
  _joystick = joy;
  displayJoystick ();
}


void ZXULA::displayJoystick () const {
  switch (_joystick) {
  case J_KEMPSTON: 
	PA_OutputSimpleText(0, 0, 20,"Joystick : Kempston    ");
    break;
  case J_SINCLAIR_1: 
	PA_OutputSimpleText(0, 0, 20,"Joystick : Sinclair 1  ");
    break;
  case J_SINCLAIR_2: 
	PA_OutputSimpleText(0, 0, 20,"Joystick : Sinclair 2  ");
    break;
  case J_CURSOR: 
	PA_OutputSimpleText(0, 0, 20,"Joystick : Cursor      ");
    break;
  default:
   break;
  }
}


// Acces en ecriture du Z80 dans la zone graphique de la memoire
void ZXULA::WriteToVideo (u16 adr, u8 data) { 
  u16 x,y;
  u8 paper, ink;
  bool blink;
  // Partie graphique 
  if (adr >= ZX_VIDEO_START && adr < ZX_ATTRIBUTE_START) {
	ZXVideoAdr2DSCoord(adr, x, y);
	GetAttributes (x, y, paper, ink, blink);
	Draw (x, y, data, paper, ink, blink);
  }
  else if (adr >= ZX_ATTRIBUTE_START && adr < ZX_ATTRIBUTE_END) { // Partie attribut
	u16 zx_video_base, zx_video_adr;
	u8 zx_video_data;
	u16 i;
	ZXAttrAdr2VideoAdr (adr, zx_video_base);
	ZXVideoAdr2DSCoord(zx_video_base, x, y);
	GetAttributes (x, y, paper, ink, blink);

	for (i=0; i<8; i++) {
	  zx_video_adr = zx_video_base + (i << 8);
	  theZXMemory->ReadFromZXMemory (zx_video_adr, zx_video_data);
	  ZXVideoAdr2DSCoord(zx_video_adr, x, y);
	  Draw (x, y, zx_video_data, paper, ink, blink);
	}
  }
}

void ZXULA::ZXAttrAdr2VideoAdr (u16 adr, u16 &v_adr) const {
  u16 ab, b, b_r;
  ab = adr - ZX_ATTRIBUTE_START;
  b = ab >> 8;
  b_r = ab & 0xFF;
  v_adr = (b << 11) + b_r + ZX_VIDEO_START;
}

void ZXULA::ZXVideoAdr2DSCoord (u16 adr, u16 &x, u16 &y) const {
  u16 y1, y1_r, y2, y2_r, y3, y3_r, ab;
  ab = adr-ZX_VIDEO_START;
  y1 = ab >> 11; // / 2048
  y1_r = ab & 0x07FF; // %2048;
  
  y2 = y1_r >> 8; // / 256;
  y2_r = y1_r & 0x00FF; // %256;
  
  y3 = y2_r >> 5; // / 32;
  y3_r = y2_r & 0x001F; // %32;
  
  y=(y1 << 6) + (y3 << 3) + y2; // y=(y1*64)+y2+(y3*8);
  x=y3_r;
}

void ZXULA::RefreshScreen (void) {
 
  // Change border color
  // No sense with DS

  // Fill the sheet
  u16 adr;
  u8 data;
  u8 paper, ink;
  bool blink;
  for (adr=ZX_VIDEO_START; adr < ZX_ATTRIBUTE_START; adr++) {
	theZXMemory->ReadFromZXMemory ((u16) adr, data);
	u16 x,y;
	ZXVideoAdr2DSCoord(adr, x, y);
	GetAttributes (x, y, paper, ink, blink);
	Draw (x, y, data, paper, ink, blink);
  }
}

void ZXULA::blink () {
  flash++;
  if (flash < 13) return;//

  flash = 0;


  if (flash_inv) flash_inv=false;
  else flash_inv=true;

  u8 paper, ink;
  bool blink;
  u16 x,y;

  for (u16 adr = ZX_ATTRIBUTE_START; adr < ZX_ATTRIBUTE_END; adr++) { 
	u16 zx_video_base, zx_video_adr;
	u8 zx_video_data;
	ZXAttrAdr2VideoAdr (adr, zx_video_base);
	ZXVideoAdr2DSCoord(zx_video_base, x, y);
	GetAttributes (x, y, paper, ink, blink);

	if (blink) {
	  for (u16 i=0; i<8; i++) {
		zx_video_adr = zx_video_base + (i << 8);
		theZXMemory->ReadFromZXMemory (zx_video_adr, zx_video_data);
		ZXVideoAdr2DSCoord(zx_video_adr, x, y);
		Draw (x, y, zx_video_data, paper, ink, blink);
	  }
	}
  }
}

  void ZXULA::GetAttributes (u8 x, u8 y, u8 &paper, u8 &ink, bool &blink) {
  // Get color attributs
  u8 attr;
  u16 att_adr = ((y << 2) & 0xFFE0) + x;
  theZXMemory->ReadFromZXMemory ((u16) (ZX_ATTRIBUTE_START + att_adr), attr); // Read color attributes

  // Extracts paper, ink
  paper = (attr >> 3) & 0x07; // D5-D3
  ink = attr & 0x07; // D2-D0
  if (attr & 0x40) { // D6 : Bright bit
    ink+=8; // Select ink color from 8 to 15 instead of 0 to 7
    paper+=8;// Select paper color from 8 to 15 instead of 0 to 7
  }

  if (attr & 0x80) blink=true;
  else blink=false;
}

// 0 <= x <= 31
// 0 <= y <= 191
void ZXULA::Draw (u16 x, u16 y, u8 data, u8 paper, u8 ink, bool blink) const {
  u8 theData = data;
  u8 c1, c2;


  if (blink == false) { // Cas normal
	c1=ink;
	c2=paper;
  }
  else { // Clignotant
	if (flash_inv) { // inversion
	  c1=paper;
	  c2=ink;
	}
	else { // normal
	  c1=ink;
	  c2=paper;
	}
  }

  for (int i=0; i < 8; i++) {
    theData = (data << i) & 0x80;
	u16 color;

    if (theData) { // ink color (foreground)
      color= ColorTable [c1];
    }
    else { // paper color (background)
      color=ColorTable [c2];
    }

	PA_Put16bitPixel(1, (x << 3)+i, y, color);
  }
}

void ZXULA::RefreshSound (void) {
  if (!_sound) return;

  // Get samples stored into the ZXIO during execution
  int nbsample = theZXIO->GetNbSamples();
  const u16 *pSampleIndex = theZXIO->GetSampleIndex ();
  const s8 *pSampleValue = theZXIO->GetSampleValue ();

  int i;
  // Fill the sound stream with previous value
  for (i=0; i < 256; i++) theSoundBuffer[i] = _lastSampleValue;
  
  // Fill and expand samples into the sound stream
  int j;
  for (j = 0; j < nbsample; j++) {
	// the Sound stream size is 256. 
	// The sample index is between 0 and 50000 (processor cycles)
	// So, index 0 correspond to 0 and sample stamped 50000 (if exists) correspond to 256
	for (i=pSampleIndex[j]/200; i < 256; i++) theSoundBuffer[i]=pSampleValue[j]; // Durty !!!
	//_lastSampleValue = aSampleValue;
  }
  _lastSampleValue = pSampleValue[nbsample-1];

  theZXIO->ClearNbSamples ();

  PA_StopSound(0); // Doit on l'arreter avant ?
  PA_PlaySound(0, theSoundBuffer, 256, 127, 11025);
}

void ZXULA::RefreshKeyboard (u8 letter) {
  u8 portF7FE;
  u8 portFBFE;
  u8 portFDFE;
  u8 portFEFE;
  u8 portEFFE;
  u8 portDFFE;
  u8 portBFFE;
  u8 port7FFE;

  // Read all
  theZXIO->ReadFromZXIO (0xF7FE, portF7FE);
  theZXIO->ReadFromZXIO (0xFBFE, portFBFE);
  theZXIO->ReadFromZXIO (0xFDFE, portFDFE);
  theZXIO->ReadFromZXIO (0xFEFE, portFEFE);
  theZXIO->ReadFromZXIO (0xEFFE, portEFFE);
  theZXIO->ReadFromZXIO (0xDFFE, portDFFE);
  theZXIO->ReadFromZXIO (0xBFFE, portBFFE);
  theZXIO->ReadFromZXIO (0x7FFE, port7FFE);

  // Left side
  if (letter==ZX_CAPS_SHIFT) portFEFE &= 0x1E; // Bit 0 // CAPS SHIFT
  else portFEFE |= 0x01; // Bit 0
  if (letter=='z') portFEFE &= 0x1D; // Bit 1
  else portFEFE |= 0x02; // Bit 1
  if (letter=='x') portFEFE &= 0x1B; // Bit 2
  else portFEFE |= 0x04; // Bit 2
  if (letter=='c') portFEFE &= 0x17; // Bit 3
  else portFEFE |= 0x08; // Bit 3
  if (letter=='v') portFEFE &= 0x0F; // Bit 4
  else portFEFE |= 0x10; // Bit 4

  if (letter=='a') portFDFE &= 0x1E; // Bit 0
  else portFDFE |= 0x01; // Bit 0
  if (letter=='s') portFDFE &= 0x1D; // Bit 1
  else portFDFE |= 0x02; // Bit 1
  if (letter=='d') portFDFE &= 0x1B; // Bit 2
  else portFDFE |= 0x04; // Bit 2
  if (letter=='f') portFDFE &= 0x17; // Bit 3
  else portFDFE |= 0x08; // Bit 3
  if (letter=='g') portFDFE &= 0x0F; // Bit 4
  else portFDFE |= 0x10; // Bit 4

  if (letter=='q') portFBFE &= 0x1E; // Bit 0
  else portFBFE |= 0x01; // Bit 0
  if (letter=='w') portFBFE &= 0x1D; // Bit 1
  else portFBFE |= 0x02; // Bit 1
  if (letter=='e') portFBFE &= 0x1B; // Bit 2
  else portFBFE |= 0x04; // Bit 2
  if (letter=='r') portFBFE &= 0x17; // Bit 3
  else portFBFE |= 0x08; // Bit 3
  if (letter=='t') portFBFE &= 0x0F; // Bit 4
  else portFBFE |= 0x10; // Bit 4

  if (letter=='1') portF7FE &= 0x1E; // Bit 0
  else portF7FE |= 0x01; // Bit 0
  if (letter=='2') portF7FE &= 0x1D; // Bit 1
  else portF7FE |= 0x02; // Bit 1
  if (letter=='3') portF7FE &= 0x1B; // Bit 2
  else portF7FE |= 0x04; // Bit 2
  if (letter=='4') portF7FE &= 0x17; // Bit 3
  else portF7FE |= 0x08; // Bit 3
  if (letter=='5') portF7FE &= 0x0F; // Bit 4
  else portF7FE |= 0x10; // Bit 4

  // Right side
  if (letter=='0') portEFFE &= 0x1E; // Bit 0
  else portEFFE |= 0x01; // Bit 0
  if (letter=='9') portEFFE &= 0x1D; // Bit 1
  else portEFFE |= 0x02; // Bit 1
  if (letter=='8') portEFFE &= 0x1B; // Bit 2
  else portEFFE |= 0x04; // Bit 2
  if (letter=='7') portEFFE &= 0x17; // Bit 3
  else portEFFE |= 0x08; // Bit 3
  if (letter=='6') portEFFE &= 0x0F; // Bit 4
  else portEFFE |= 0x10; // Bit 4

  if (letter=='p') portDFFE &= 0x1E; // Bit 0
  else portDFFE |= 0x01; // Bit 0
  if (letter=='o') portDFFE &= 0x1D; // Bit 1
  else portDFFE |= 0x02; // Bit 1
  if (letter=='i') portDFFE &= 0x1B; // Bit 2
  else portDFFE |= 0x04; // Bit 2
  if (letter=='u') portDFFE &= 0x17; // Bit 3
  else portDFFE |= 0x08; // Bit 3
  if (letter=='y') portDFFE &= 0x0F; // Bit 4
  else portDFFE |= 0x10; // Bit 4

  if (letter=='\n') portBFFE &= 0x1E; // Bit 0
  else portBFFE |= 0x01; // Bit 0
  if (letter=='l') portBFFE &= 0x1D; // Bit 1
  else portBFFE |= 0x02; // Bit 1
  if (letter=='k') portBFFE &= 0x1B; // Bit 2
  else portBFFE |= 0x04; // Bit 2
  if (letter=='j') portBFFE &= 0x17; // Bit 3
  else portBFFE |= 0x08; // Bit 3
  if (letter=='h') portBFFE &= 0x0F; // Bit 4
  else portBFFE |= 0x10; // Bit 4

  if (letter==' ') port7FFE &= 0x1E; // Bit 0
  else port7FFE |= 0x01; // Bit 0
  if (letter==ZX_SYMBOL_SHIFT) port7FFE &= 0x1D; // Bit 1 // SYMBOL SHIFT
  else port7FFE |= 0x02; // Bit 1
  if (letter=='m') port7FFE &= 0x1B; // Bit 2
  else port7FFE |= 0x04; // Bit 2
  if (letter=='n') port7FFE &= 0x17; // Bit 3
  else port7FFE |= 0x08; // Bit 3
  if (letter=='b') port7FFE &= 0x0F; // Bit 4
  else port7FFE |= 0x10; // Bit 4

  // Hack for BackSpace/Up/Down/Left/Right
  // Mapping for the arrows
  // 0 : None
  // 2 : Sinclair 2 interface Joystick 1
  // 3 : Sinclair 2 interface Joystick 2
  // 4 : Cursor interface
  // 5 : Kempston interface

  if (_joystick == J_SINCLAIR_1) { // Sinclair 2 interface : Joystick 1
    if (Pad.Held.Left) portEFFE &= 0xEF; // Bit 4
    if (Pad.Held.Right) portEFFE &= 0xF7; // Bit 3
    if (Pad.Held.Down) portEFFE &= 0xFB; // Bit 2
    if (Pad.Held.Up) portEFFE &= 0xFD; // Bit 1
    if (Pad.Held.A) portEFFE &= 0xFE; // Bit 0
  }
  else if (_joystick == J_SINCLAIR_2) { // Sinclair 2 interface : Joystick 2
    if (Pad.Held.Left) portF7FE &= 0xFE; // Bit 0
    if (Pad.Held.Right) portF7FE &= 0xFD;// Bit 1
    if (Pad.Held.Down) portF7FE &= 0xFB; // Bit 2
    if (Pad.Held.Up) portF7FE &=  0xF7; // Bit 3
    if (Pad.Held.A) portF7FE &=0xEF; // Bit 4
  }
  else if (_joystick == J_CURSOR) { // Cursor interface
    if (Pad.Held.Left) portF7FE &= 0x0F;
    if (Pad.Held.Right) portEFFE &= 0x1B;
    if (Pad.Held.Up) portEFFE &= 0x17;
    if (Pad.Held.Down) portEFFE &= 0x0F;
    if (Pad.Held.A) portEFFE &= 0x1E;
  }
  else if (_joystick == J_KEMPSTON) { // Kempston interface
     u8 port001F; // Kempston port
     theZXIO->ReadFromZXIO (0x001F, port001F);
  
     if (Pad.Held.Left) port001F |= 0x02; // Bit 1
     else port001F &= 0xFD;
     if (Pad.Held.Right) port001F |=0x01; // Bit 0
     else port001F &= 0xFE;
     if (Pad.Held.Up) port001F |= 0x08; // Bit 3
     else port001F &= 0xF7;
     if (Pad.Held.Down) port001F |= 0x04; // Bit 2
     else port001F &= 0xFB;
     if (Pad.Held.A)port001F  |= 0x10; // Bit 4
     else port001F &= 0xEF;
   
     theZXIO->WriteToZXIO (0x001F, port001F);
  }

  // Write all
  theZXIO->WriteToZXIO (0xF7FE, portF7FE);
  theZXIO->WriteToZXIO (0xFBFE, portFBFE);
  theZXIO->WriteToZXIO (0xFDFE, portFDFE);
  theZXIO->WriteToZXIO (0xFEFE, portFEFE);
  theZXIO->WriteToZXIO (0xEFFE, portEFFE);
  theZXIO->WriteToZXIO (0xDFFE, portDFFE);
  theZXIO->WriteToZXIO (0xBFFE, portBFFE);
  theZXIO->WriteToZXIO (0x7FFE, port7FFE);
}

int ZXULA::Screenshot (const char *file) const {
  return 0;
}

//
// ZXIO
//
ZXIO::ZXIO (void) {
  // A map could be better than an array ...
  io = new u8 [65536];
  for (int i = 0; i < 65536; i++) {
    io [i] = 0xFF;
  }

  // Kempston port
  io [0x001F] = 0;
  _nSamples=0;
}

ZXIO::~ZXIO (void) {
  if (io) {
    delete [] io;
    io = 0;
  }
}

int ZXIO::CPUWriteToZXIO (u16 Port, u8 data, int current_cycle) {
#if 0
  switch (Port) {
  case 65278:
  case 65022:
  case 64510:
  case 63486:
  case 61438:
  case 57342:
  case 49150:
  case 32766:
    break;
  default:
    printf ("ZXIO::WriteToZXIO : Port %04X, Data %02X\n", Port, data);
    break;
  }
  printf ("ZXIO::CPUWriteToZXIO : Port %04X, Data %02X\n", Port, data);
#endif
  // Manage port 254
  // D2-D1-D0 : border color
  // D3 : MIC
  // D4 : Speaker
  if ((Port & 0xE0FE) == 0x00FE) { // A7-A5 must be stucked at 0 if it is not the keyboard
    // border_color = data & 0x07; // D2-D1-D0
    // mic = data & 0x08; // D3

	s8 speaker; // Bit D4

    if (data & 0x10) speaker = 127; // D4
    else speaker = -126;

	// On ajoute le sample
	_SampleIndex [_nSamples] = 50000-current_cycle; // 50000 cycles en 16ms
	_SampleValue [_nSamples] = speaker;
	_nSamples++;
	printf ("Sample:%d N=%d\n", speaker, _nSamples);
  }

  io [Port] = data; // usefull ?
  return 0;
}

int ZXIO::WriteToZXIO (u16 Port, u8 data) {
  io [Port] = data;
  return 0;
}

int ZXIO::CPUReadFromZXIO (u16 Port, u8& data) const {
#if 0
  switch (Port) {
  case 65278:
  case 65022:
  case 64510:
  case 63486:
  case 61438:
  case 57342:
  case 49150:
  case 32766:
    break;
  default:
    printf ("ZXIO::CPUReadFromZXIO : Port %04X, Data %02X\n", Port, data);
    break;
  }
#endif

  // hack for Kempston. Sometimes port is 001F (SABRE) and sometimes 021F (PSSST)
  if (Port & 0x001F == 0x001F) data = io [0x001F];
  else data = io [Port];
  return 0;
}

int ZXIO::ReadFromZXIO (u16 Port, u8& data) const {
  data = io [Port];
#if 0
  switch (Port) {
  case 65278:
  case 65022:
  case 64510:
  case 63486:
  case 61438:
  case 57342:
  case 49150:
  case 32766:
    break;
  default:
    printf ("ZXIO::ReadFromZXIO : Port %04X, Data %02X\n", Port, data);
    break;
  }
#endif

  return 0;
}

//
// SPECTRUM
//
ZXSpectrum::ZXSpectrum (const char* aRomFile) {
  theZXMemory = new ZXMemory ();
  theZXIO = new ZXIO ();
  theZXULA = new ZXULA (theZXMemory, theZXIO);
  theProcessor = new Z80 (theZXMemory, theZXULA, theZXIO);

  bool res = theZXMemory->LoadFromFile (aRomFile); // ROM at @0000
  
  if (res == false) { // On charge la rom depuis le binaire
	theZXMemory->LoadFromBin (rom);
  }

  theKeyboard = new ZXKeyboard();
  theMenu = new ZXMenu(theZXULA);
  theSequencer = new ZXSequencer();
}

ZXSpectrum::~ZXSpectrum (void) {
  if (theZXMemory) {
    delete theZXMemory;
    theZXMemory = 0;
  }

  if (theZXULA) {
    delete theZXULA;
    theZXULA = 0;
  }

  if (theZXIO) {
    delete theZXIO;
    theZXIO = 0;
  }

  if (theProcessor) {
    delete theProcessor;
    theProcessor = 0;
  }

  if (theKeyboard) {
	delete theKeyboard;
	theKeyboard = 0;
  }

  if (theMenu) {
	delete theMenu;
	theMenu=0;
  }

  if (theSequencer) {
	delete theSequencer;
	theSequencer=0;
  }
}

//
// Menu
//
int ZXSpectrum::Run (void) {
  string snapshot;
  deque<string> sequence;

  // Menu initial
  theMenu->show();
  if (theMenu->process (snapshot, sequence) == 0) {
	LoadSNAFile (snapshot.c_str());
	theSequencer->setSequence (sequence);
  }
  theMenu->hide();
  theKeyboard->show();
  theZXULA->RefreshScreen();

  // Le vrai run
  while (1) {

	if (Pad.Newpress.Start) {
	  theKeyboard->hide();
	  theMenu->show();
	  if (theMenu->process (snapshot, sequence) == 0) {
		LoadSNAFile (snapshot.c_str());
		theZXULA->RefreshScreen();
		theSequencer->setSequence (sequence);
	  }
	  theMenu->hide();
	  theKeyboard->show();
    }

	if (Pad.Newpress.Select) {
	  theZXULA->changeJoystick();
	}

    // Run
	theProcessor->Execute (50000);
	Keyboard ();
	theProcessor->Z80_Cause_Interrupt (0);
	Sound ();
	if (Pad.Newpress.R) PA_OutputText(0, 0, 21, "FPS : %02d", PA_RTC.FPS);  // Display the frames per second
	if (Pad.Released.R) PA_OutputText(0, 0, 21, "          ");
	theZXULA->blink();
	PA_WaitForVBL();
  }

  return 0;
}

void ZXSpectrum::Keyboard () {
  char k=theSequencer->getKey();
  if (!k) {
	k = theKeyboard->getKey();
  }
  theZXULA->RefreshKeyboard (k);
}

void ZXSpectrum::Sound () {
  theZXULA->RefreshSound ();
}

void ZXSpectrum::Reset (void) {
  theProcessor->Reset ();
}

/*
 *      Load a 48K  .SNA file.
 *
 *      48K Format as follows:
 *      Offset  Size    Description (all registers stored with LSB first)
 *      0       1       I
 *      1       18      HL',DE',BC',AF',HL,DE,BC,IY,IX
 *      19      1       Interrupt (bit 2 contains IFF2 1=EI/0=DI
 *      20      1       R
 *      21      4       AF,SP
 *      25      1       Interrupt Mode (0=IM0/1=IM1/2=IM2)
 *      26      1       Border Colour (0..7)
 *      27      48K     RAM dump 0x4000-0xffff
 *      PC is stored on stack. */

int ZXSpectrum::LoadSNAFile (const char *aSNAFile) {
  FILE *fp = fopen (aSNAFile, "rb");
  if (!fp) {
    printf ("ZXSpectrum::LoadSNAFile : unable to open SNA file %s\n", aSNAFile);
    return 1;
  }

  Z80_Regs regs;
  u8 dummy;

  fread (&regs.I, sizeof (u8), 1, fp); // I
  regs.I &= 0xFF;

  fread (&regs.HL2.B.l, sizeof (u8), 1, fp); // HL'
  fread (&regs.HL2.B.h, sizeof (u8), 1, fp);
  regs.HL2.B.l &= 0xFF;
  regs.HL2.B.h &= 0xFF;

  fread (&regs.DE2.B.l, sizeof (u8), 1, fp); // DE'
  fread (&regs.DE2.B.h, sizeof (u8), 1, fp);
  regs.DE2.B.l &= 0xFF;
  regs.DE2.B.h &= 0xFF;

  fread (&regs.BC2.B.l, sizeof (u8), 1, fp); // BC'
  fread (&regs.BC2.B.h, sizeof (u8), 1, fp);
  regs.BC2.B.l &= 0xFF;
  regs.BC2.B.h &= 0xFF;

  fread (&regs.AF2.B.l, sizeof (u8), 1, fp); // AF'
  fread (&regs.AF2.B.h, sizeof (u8), 1, fp);
  regs.AF2.B.l &= 0xFF;
  regs.AF2.B.h &= 0xFF;

  fread (&regs.HL.B.l, sizeof (u8), 1, fp); // HL
  fread (&regs.HL.B.h, sizeof (u8), 1, fp);
  regs.HL.B.l &= 0xFF;
  regs.HL.B.h &= 0xFF;

  fread (&regs.DE.B.l, sizeof (u8), 1, fp); // DE
  fread (&regs.DE.B.h, sizeof (u8), 1, fp);
  regs.DE.B.l &= 0xFF;
  regs.DE.B.h &= 0xFF;

  fread (&regs.BC.B.l, sizeof (u8), 1, fp); // BC
  fread (&regs.BC.B.h, sizeof (u8), 1, fp);
  regs.BC.B.l &= 0xFF;
  regs.BC.B.h &= 0xFF;

  fread (&regs.IY.B.l, sizeof (u8), 1, fp); // IY
  fread (&regs.IY.B.h, sizeof (u8), 1, fp);
  regs.IY.B.l &= 0xFF;
  regs.IY.B.h &= 0xFF;

  fread (&regs.IX.B.l, sizeof (u8), 1, fp); // IX
  fread (&regs.IX.B.h, sizeof (u8), 1, fp);
  regs.IX.B.l &= 0xFF;
  regs.IX.B.h &= 0xFF;

  fread (&dummy, sizeof (u8), 1, fp); // IFF2
  regs.IFF2 = (dummy & 0x04) >> 2;
  regs.IFF1 = (dummy & 0x04) >> 2;

  fread (&dummy, sizeof (u8), 1, fp); // R
  regs.R = dummy & 127;
  regs.R2 = dummy & 128;

  fread (&regs.AF.B.l, sizeof (u8), 1, fp); // AF
  fread (&regs.AF.B.h, sizeof (u8), 1, fp);
  regs.AF.B.l &= 0xFF;
  regs.AF.B.h &= 0xFF;

  fread (&regs.SP.B.l, sizeof (u8), 1, fp); // SP
  fread (&regs.SP.B.h, sizeof (u8), 1, fp);
  regs.SP.B.l &= 0xFF;
  regs.SP.B.h &= 0xFF;

  fread (&regs.IM, sizeof (u8), 1, fp); // IM
  regs.IM &= 0xFF;

  fread (&dummy, sizeof (u8), 1, fp); // Border color
  //theZXIO->SetBorderColor (dummy & 0x07);

  // Fill the Memory
  int address = 0x4000;
  u8 data;
  while (fread (&data, sizeof (u8), 1, fp) != 0) {
	if (address>=65536) {
      fclose (fp);
      return 1;
	}

    theZXMemory->WriteToZXMemory ((u16)address, data);
    address++;
  }

  /* get pc from stack */
  address = regs.SP.W.l & 0xFFFF; // Get SP
  theZXMemory->ReadFromZXMemory ((u16) address, regs.PC.B.l);
  theZXMemory->ReadFromZXMemory ((u16) (address+1), regs.PC.B.h);
    
  address += 2;
  regs.SP.W.l = address & 0xFFFF; // New stack pointer

  regs.pending_irq = 0;
  regs.pending_nmi = 0;

  // Set the real Registers
  theProcessor->Z80_SetRegs (regs);
  fclose (fp);
  return 0; 
}


/*
 *      Save a 48K  .SNA file.
 *
 *      48K Format as follows:
 *      Offset  Size    Description (all registers stored with LSB first)
 *      0       1       I
 *      1       18      HL',DE',BC',AF',HL,DE,BC,IY,IX
 *      19      1       Interrupt (bit 2 contains IFF2 1=EI/0=DI
 *      20      1       R
 *      21      4       AF,SP
 *      25      1       Interrupt Mode (0=IM0/1=IM1/2=IM2)
 *      26      1       Border Colour (0..7)
 *      27      48K     RAM dump 0x4000-0xffff
 *      PC is stored on stack. */

int ZXSpectrum::SaveSNAFile (const char *aSNAFile) const {
  FILE *fp = fopen (aSNAFile, "wb");
  if (!fp) {
    printf ("ZXSpectrum::SaveSNAFile : unable to open SNA file %s\n", aSNAFile);
    return 1;
  }

  Z80_Regs regs= theProcessor->Z80_GetRegs ();
  u8 dummy;

  /* set pc to stack */
  u16 adr;
  adr = regs.SP.W.l & 0xFFFF; // Get SP
  theZXMemory->WriteToZXMemory ((u16) (adr-2), regs.PC.B.l);
  theZXMemory->WriteToZXMemory ((u16) (adr-1), regs.PC.B.h);
  adr -= 2;
  regs.SP.W.l = adr & 0xFFFF; // New stack pointer

  regs.I &= 0xFF;
  fwrite (&regs.I, sizeof (u8), 1, fp); // I

  regs.HL2.B.l &= 0xFF;
  fwrite (&regs.HL2.B.l, sizeof (u8), 1, fp); // HL'
  regs.HL2.B.h &= 0xFF;
  fwrite (&regs.HL2.B.h, sizeof (u8), 1, fp);

  regs.DE2.B.l &= 0xFF;
  regs.DE2.B.h &= 0xFF;
  fwrite (&regs.DE2.B.l, sizeof (u8), 1, fp); // DE'
  fwrite (&regs.DE2.B.h, sizeof (u8), 1, fp);

  regs.BC2.B.l &= 0xFF;
  regs.BC2.B.h &= 0xFF;
  fwrite (&regs.BC2.B.l, sizeof (u8), 1, fp); // BC'
  fwrite (&regs.BC2.B.h, sizeof (u8), 1, fp);

  regs.AF2.B.l &= 0xFF;
  regs.AF2.B.h &= 0xFF;
  fwrite (&regs.AF2.B.l, sizeof (u8), 1, fp); // AF'
  fwrite (&regs.AF2.B.h, sizeof (u8), 1, fp);

  regs.HL.B.l &= 0xFF;
  regs.HL.B.h &= 0xFF;
  fwrite (&regs.HL.B.l, sizeof (u8), 1, fp); // HL
  fwrite (&regs.HL.B.h, sizeof (u8), 1, fp);

  regs.DE.B.l &= 0xFF;
  regs.DE.B.h &= 0xFF;
  fwrite (&regs.DE.B.l, sizeof (u8), 1, fp); // DE
  fwrite (&regs.DE.B.h, sizeof (u8), 1, fp);

  regs.BC.B.l &= 0xFF;
  regs.BC.B.h &= 0xFF;
  fwrite (&regs.BC.B.l, sizeof (u8), 1, fp); // BC
  fwrite (&regs.BC.B.h, sizeof (u8), 1, fp);
  
  regs.IY.B.l &= 0xFF;
  regs.IY.B.h &= 0xFF;
  fwrite (&regs.IY.B.l, sizeof (u8), 1, fp); // IY
  fwrite (&regs.IY.B.h, sizeof (u8), 1, fp);

  regs.IX.B.l &= 0xFF;
  regs.IX.B.h &= 0xFF;
  fwrite (&regs.IX.B.l, sizeof (u8), 1, fp); // IX
  fwrite (&regs.IX.B.h, sizeof (u8), 1, fp);

  dummy = (regs.IFF1 & 0x01) << 2;
  fwrite (&dummy, sizeof (u8), 1, fp); // IFF2

  dummy = (regs.R & 127) | (regs.R2 & 128);
  fwrite (&dummy, sizeof (u8), 1, fp); // R

  regs.AF.B.l &= 0xFF;
  regs.AF.B.h &= 0xFF;
  fwrite (&regs.AF.B.l, sizeof (u8), 1, fp); // AF
  fwrite (&regs.AF.B.h, sizeof (u8), 1, fp);
  
  regs.SP.B.l &= 0xFF;
  regs.SP.B.h &= 0xFF;
  fwrite (&regs.SP.B.l, sizeof (u8), 1, fp); // SP
  fwrite (&regs.SP.B.h, sizeof (u8), 1, fp);
 
  regs.IM &= 0xFF;
  fwrite (&regs.IM, sizeof (u8), 1, fp); // IM
 
  //dummy = theZXIO->GetBorderColor ();
  dummy = 0;
  fwrite (&dummy, sizeof (u8), 1, fp); // Border color

  // Dump the Memory
  u8 data;
  for (int address = 0x4000; address <= 0xFFFF; address++) {
    theZXMemory->ReadFromZXMemory ((u16) address, data);
    fwrite (&data, sizeof (u8), 1, fp);
  }

  fclose (fp);

  return 0; 
}

int main (int argc, char **argv) { 

  // Graphical init
  PA_Init();    // Initializes PA_Lib
  PA_InitVBL(); // Initializes a standard VBL
  PA_InitSound();  // Init the sound system

  // Splash
  //PA_SplashBlue();
  //PA_SplashEyes();

  PA_Init16bitBg(1, 3);
  PA_Init16bitBg(0, 3);

  PA_InitText(0, 0);  	// Background number, from 0-3, on which to load the text system
  PA_SetTextCol(0, 31, 31, 31);
  PA_Clear16bitBg(0);

  /* install a digital sound driver */

  // FAT init
  fatInitDefault();

#if 0
  char buf[64];
  getcwd (buf, 63);
  PA_OutputText(0, 0, 21, "getcwd : %s", buf);
#endif

  // Create and launch the simulation
  ZXSpectrum aSpectrum ("spectrum.rom");
  aSpectrum.Run ();
  while (1) { // On ne doit pas sortir
	PA_WaitForVBL();
  }

  return 0;
}

void ZXSpectrum::DumpRegs() {
#ifdef DEBUG
  const Z80_Regs & r = theProcessor->Z80_GetRegs();
  PA_Print (0, "A:%02d\n", r.AF.B.l);
  PA_Print (0, "B:%02d\n", r.BC.B.l);
  PA_Print (0, "C:%02d\n", r.BC.B.h);
  PA_Print (0, "D:%02d\n", r.DE.B.l);
  PA_Print (0, "E:%02d\n", r.DE.B.h);
  PA_Print (0, "PC:%04d\n", r.PC.W.l);
#endif
}

void ZXSpectrum::TestAttVideo () { 
  u16 adr;
  u8 data;
  u8 i;

  theZXMemory->WriteToZXMemory (ZX_ATTRIBUTE_START, 0x0E);
  theZXMemory->WriteToZXMemory (ZX_ATTRIBUTE_START+1, 0x1C);
  adr = ZX_VIDEO_START;
  data = 0xF0;
  for (i=0; i < 8; i++) {
	theZXMemory->WriteToZXMemory (adr, data);
	theZXULA->WriteToVideo (adr, data);
	adr = adr + 256;
  }
  adr = ZX_VIDEO_START+1;
  data = 0x3C;
  for (i=0; i < 8; i++) {
	theZXMemory->WriteToZXMemory (adr, data);
	theZXULA->WriteToVideo (adr, data);
	adr = adr + 256;
  }
  theZXMemory->WriteToZXMemory (ZX_ATTRIBUTE_START, 0x14);
  theZXULA->WriteToVideo (ZX_ATTRIBUTE_START, 0x14);
}


