/*
 *  Copyright (C) 2008 Florent Bedoiseau (electronique.fb@free.fr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef ZXSPECTRUM_H
#define ZXSPECTRUM_H

#include "Z80.h"
#include "rom.h"

#define SCREEN_W 256
#define SCREEN_H 192

#define ZX_VIDEO_START 16384
#define ZX_ATTRIBUTE_START 22528
#define ZX_ATTRIBUTE_END 23296

class ZXKeyboard;
class ZXMenu;
class ZXSequencer;

#include "ZXCfg.h"

class ZXMemory {
 public:
  ZXMemory ();
  ~ZXMemory (void);

  bool LoadFromFile (const char* aFile);
  bool LoadFromBin (const u8* aBin);
  bool SaveToFile (const char* aFile);

  inline void WriteToZXMemory (u16 adr, u8 data) { memory [adr] = data; }
  inline void ReadFromZXMemory (u16 adr, u8& data) const { data = memory [adr]; }

 private:
  u8 *memory;
};

class ZXULA {
 public:
  ZXULA (ZXMemory *aZXMemory, ZXIO *aZXIO);
  ~ZXULA (void);
  void RefreshScreen (void);
  void RefreshKeyboard (u8);
  void RefreshSound (void);
  int Screenshot (const char *file) const;
  void setJoystick (ZX_JOYSTICK joy);
  void setSound (bool sound) { _sound = sound; }
  bool getSound () const { return _sound; }
  void changeJoystick ();
  void blink ();
  void WriteToVideo (u16 adr, u8 data);
 private:
  inline void Draw (u16 x, u16 y, u8 data, u8 paper, u8 ink, bool inv) const;
  inline void ZXVideoAdr2DSCoord (u16 adr, u16 &x, u16 &y) const;
  inline void ZXAttrAdr2VideoAdr (u16 adr, u16 &v_adr) const;

  void GetAttributes (u8 x, u8 y, u8 &paper, u8 &ink, bool &blink);
  void displayJoystick () const;

  int ColorTable [16];

  int flash;
  bool flash_inv;

  ZX_JOYSTICK _joystick;

  ZXMemory *theZXMemory;
  ZXIO *theZXIO;

  // Sound
  bool _sound;
  s8 _lastSampleValue;
  s8 *theSoundBuffer;
};

class ZXIO {
 public:
  ZXIO (void);
  ~ZXIO (void);

  // Access to IO without time-stamp
  int WriteToZXIO (u16 Port, u8 data);
  int ReadFromZXIO (u16 Port, u8& data) const;

  // Access to IO with time-stamp
  int CPUWriteToZXIO (u16 Port, u8 data, int current_cycle);
  int CPUReadFromZXIO (u16 Port, u8& data) const;

  int GetNbSamples () const { return _nSamples; }
  void ClearNbSamples () { _nSamples=0; }
  const u16 *GetSampleIndex () const { return _SampleIndex; }
  const s8 *GetSampleValue () const { return _SampleValue; }

 private:
  u8 *io;

  // Buffer de samples
  u16 _SampleIndex [300];
  s8 _SampleValue [300];
  int _nSamples; // 48 samples pour Sabre Wulf, 99 pour Project Futur, 112 JetPac
};

class ZXSpectrum {
 public:
  ZXSpectrum (const char* aRomFile);
  ~ZXSpectrum (void);
  int LoadSNAFile (const char *aSNAFile);
  int SaveSNAFile (const char *aSNAFile) const;

  int Run (void);
  void Reset (void);

  // DEBUG
  void DumpRegs();
  void TestAttVideo();

 private:
  bool _sound;

  void Keyboard (void);
  void Sound (void);
  

  Z80 *theProcessor;
  ZXMemory *theZXMemory;
  ZXULA *theZXULA;
  ZXIO *theZXIO;
  ZXKeyboard *theKeyboard;
  ZXMenu *theMenu;
  ZXSequencer *theSequencer;
};


#endif
